#include <stdlib.h>
#include <assert.h>
#include <stdio.h>
#include <err.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <stdint.h>
#include <string.h>
#include <inttypes.h>
#include <stdbool.h>
#include "apfs.h"

#pragma clang diagnostic ignored "-Waddress-of-packed-member"

#define MAX_MAPPINGS 1000000

static char *img;
static size_t img_size;
static nx_superblock_t *sb;
static omap_phys_t *sb_object_map, *fs_object_map;

static void *blockref_(uint64_t idx, uint32_t expected_type) {
  if ((idx != 0 || sb != NULL) && idx >= sb->nx_block_count)
    errx(1, "bad block access (type 0x%x): index 0x%" PRIx64 " is beyond the last block", expected_type, idx);
  uint64_t offset = (idx == 0) ? 0 : (idx * sb->nx_block_size);
  void *obj = img + offset;
  obj_phys_t *obj_hdr = obj;
  uint32_t actual_type = obj_hdr->o_type & OBJECT_TYPE_MASK;
  if (actual_type != expected_type && expected_type != OBJECT_TYPE_INVALID)
    errx(1, "bad block access (index 0x%"PRIx64"): expected type 0x%x, but got 0x%x", idx, expected_type, actual_type);
  return obj;
}
#define BLOCKREF(idx_, typenum) blockref_((idx_), (typenum))

static btree_info_t *btree_info_by_root(btree_node_phys_t *root) {
  assert((root->btn_o.o_type & OBJECT_TYPE_MASK) == OBJECT_TYPE_BTREE);
  return (btree_info_t*) (((char*)root)+sb->nx_block_size-sizeof(btree_info_t));
}

struct buf {
  void *p;
  size_t l;
};
struct kv_bufs {
  struct buf k, v;
};
static struct kv_bufs btree_node_get_kv_bufs(int idx, btree_node_phys_t *node, btree_node_phys_t *root) {
  if (idx >= node->btn_nkeys)
    errx(1, "bad btree key index");
  kvloc_t locs;
  btree_info_fixed_t *info = &btree_info_by_root(root)->bt_fixed;
  if (node->btn_flags & BTNODE_FIXED_KV_SIZE) {
    kvoff_t offsets = ((kvoff_t *)(node->btn_data))[idx];
    locs = (kvloc_t){
      .k = { .off = offsets.k, .len = info->bt_key_size },
      .v = { .off = offsets.v, .len = info->bt_val_size }
    };
  } else {
    locs = ((kvloc_t *)(node->btn_data))[idx];
  }
  char *keys_start = (char*)node->btn_data + node->btn_table_space.len;
  char *values_end = (char*)node + info->bt_node_size - ((node->btn_flags & BTNODE_ROOT) ? sizeof(btree_info_t) : 0);
  return (struct kv_bufs) {
    .k = { .p = keys_start + locs.k.off, .l = locs.k.len },
    .v = { .p = values_end - locs.v.off, .l = locs.v.len }
  };
}

static omap_val_t omap_lookup_(omap_phys_t *omap, oid_t oid, btree_node_phys_t *root, btree_node_phys_t *tree_node) {
  //printf("omap_lookup_() [fixed=%d]\n", (int)!!(tree_node->btn_flags & BTNODE_FIXED_KV_SIZE));
  //if (!(tree_node->btn_flags & BTNODE_LEAF)) errx(1, "omap isn't small");
  omap_val_t *res = NULL;
  oid_t matching_child = 0;
  for (int i=0; i<tree_node->btn_nkeys; i++) {
    struct kv_bufs bufs = btree_node_get_kv_bufs(i, tree_node, root);
    if (bufs.k.l != sizeof(omap_key_t)) errx(1, "bad size 1");
    omap_key_t *k = (omap_key_t *)bufs.k.p;
   
    if (tree_node->btn_flags & BTNODE_LEAF) {
      if (bufs.v.l != sizeof(omap_val_t)) errx(1, "bad size 2");
      omap_val_t *v = (omap_val_t *)bufs.v.p;
      //printf("entry %d: (oid=0x%"PRIx64", xid=0x%"PRIx64") -> %p=(flags=0x%x, size=0x%x, paddr=0x%"PRIx64")\n", i, k->ok_oid, k->ok_xid, v, v->ov_flags, v->ov_size, v->ov_paddr);
      if (k->ok_oid == oid) {
        if (res != NULL) errx(1, "duplicate oid 0x%"PRIx64, oid);
        res = v;
      }
    } else {
      //if (bufs.v.l != sizeof(oid_t)) warnx("bad size 3 (got %d, wanted %d)", (int)bufs.v.l, (int)sizeof(paddr_t));
      //printf("t-entry %d: (oid=0x%"PRIx64", xid=0x%"PRIx64") -> 0x%"PRIx64"\n", i, k->ok_oid, k->ok_xid, *(oid_t*)bufs.v.p);
      if (k->ok_oid > oid) {
        //printf("GT ENTRY HIT (0x%"PRIx64" < 0x%"PRIx64")\n", k->ok_oid, oid);
        break;
      }
      matching_child = *(oid_t*)bufs.v.p;
    }
  }
  if ((tree_node->btn_flags & BTNODE_LEAF) == 0) {
    if (matching_child == 0) errx(1, "omap_lookup_ found nothing");
    btree_node_phys_t *child_node = BLOCKREF(matching_child, OBJECT_TYPE_BTREE_NODE);
    return omap_lookup_(omap, oid, root, child_node);
  }
  if (res == NULL) errx(1, "unable to find oid 0x%"PRIx64, oid);
  return *res;
}

static omap_val_t omap_lookup(omap_phys_t *omap, oid_t oid) {
  //printf("omap_lookup(0x%lx)\n", oid);
  btree_node_phys_t *root = BLOCKREF(omap->om_tree_oid, OBJECT_TYPE_BTREE);
  return omap_lookup_(omap, oid, root, root);
}

/*
 * [LaTeX syntax comment]
 * The fletcher64 checksum of N elements is basically:
 * $$low\_bits_N = \sum_{i=1}^{N} data_{i} \pmod{2^{32}-1}$$
 * $$high\_bits_N = \sum_{i=1}^{N} data_{i} (N - i + 1) \pmod{2^{32}-1}$$
 * However, APFS doesn't directly store this checksum; instead, it requires that
 * the checksum of the normal checksummed data ***followed by*** the
 * two-uint32_t "checksum field" (which is actually stored in front of that
 * data) is zero.
 * This means that we have to solve for:
 * $$0 = low\_bits_N + c_1 + c_2 \pmod{2^{32}-1}$$
 * $$0 = high\_bits_N + 2 low\_bits_N + 2 c_1 + c_2 \pmod{2^{32}-1}$$
 * which has the solution:
 * $$c_1 = -(low\_bits_N + high\_bits_N) \pmod{2^{32}-1}$$
 * $$c_2 = high\_bits_N \pmod{2^{32}-1}$$
 */
static uint64_t fletcher64(void *buf, size_t len) {
  uint64_t a = 0, b = 0;
  assert(len % 4 == 0);
  for (int i=0; i<len; i+=4) {
    uint32_t n;
    memcpy(&n, (char*)buf+i, 4);
    a = (a + n) % 0xffffffffUL;
    b = (b + a) % 0xffffffffUL;
  }
  return a | (b << 32);
}
static uint64_t apfs_checksum_postprocess(uint64_t in) {
  uint64_t in_upper = in >> 32;
  uint64_t in_lower = in & 0xffffffffUL;

  // out_lower = -(in_upper + in_lower) (mod 2^32-1)
  // ["-x mod N" is "(N-x)%N"]
  uint32_t out_lower = (0xffffffffUL - ((in_upper + in_lower) % 0xffffffffUL)) % 0xffffffffUL;

  // out_upper = in_upper
  return (in_upper << 32) | out_lower;
}

struct ino_mapping { uint64_t old; uint64_t new; };
int inode_pairs;
struct ino_mapping *ino_map;

// track whether we've started renumbering stuff.
// because we don't have logic for proper btree insert/remove,
// we need to renumber a contiguous set of inodes at the end
// of the btree; otherwise the keys won't be sorted anymore.
static bool renumbering = false;

bool remap_inode(uint64_t *p, bool is_key) {
  for (int i=0; i<inode_pairs; i++) {
    struct ino_mapping *m = &ino_map[i];
    if ((*p & OBJ_ID_MASK) == m->old) {
      uint64_t new_val = (*p & OBJ_TYPE_MASK) | m->new;
      printf("  **modifying**: %"PRIx64" -> %"PRIx64"\n", *p, new_val);
      *p = new_val;
      if (is_key) renumbering = true;
      return true;
    }
  }
  if (is_key && renumbering) errx(1, "un-renumbered key after renumbered key");
  return false;
}

void process_fstree_node(btree_node_phys_t *tree_node, btree_node_phys_t *tree_root) {
  printf("<process_fstree_node() entry>\n");
  if (tree_node->btn_o.checksum_number != apfs_checksum_postprocess(fletcher64((char*)tree_node+MAX_CKSUM_SIZE, sb->nx_block_size-MAX_CKSUM_SIZE)))
    errx(1, "fstree has bad checksum");
  for (int i=0; i<tree_node->btn_nkeys; i++) {
    struct kv_bufs bufs = btree_node_get_kv_bufs(i, tree_node, tree_root);
    if (bufs.k.l < sizeof(j_key_t)) errx(1, "bad size");
    j_key_t *k_base = bufs.k.p;
    if (tree_node->btn_flags & BTNODE_LEAF) {
      printf("object 0x%"PRIx64" (%s)\n", k_base->obj_id_and_type, apfs_type_names[k_base->obj_id_and_type >> OBJ_TYPE_SHIFT]);
      remap_inode(&k_base->obj_id_and_type, true);
      switch (k_base->obj_id_and_type >> OBJ_TYPE_SHIFT) {
      case APFS_TYPE_DIR_REC:;
        j_drec_hashed_key_t *drec_key = bufs.k.p;
        j_drec_val_t *drec_val = bufs.v.p;
        printf("  '%s' -> 0x%"PRIx64"\n", drec_key->name, drec_val->file_id);
        remap_inode(&drec_val->file_id, false);
        break;
      case APFS_TYPE_INODE:;
        j_inode_val_t *ino_val = bufs.v.p;
        printf("  parent=0x%"PRIx64", private=0x%"PRIx64"\n", ino_val->parent_id, ino_val->private_id);
        remap_inode(&ino_val->private_id, false);
        break;
      }
    } else {
      if (bufs.v.l != 8) errx(1, "non-root btree node with bogus len");
      uint64_t child_loc = *(uint64_t*)bufs.v.p;
      printf("subtree 0x%"PRIx64" (%s), child_loc=0x%llx\n", k_base->obj_id_and_type, apfs_type_names[k_base->obj_id_and_type >> OBJ_TYPE_SHIFT], child_loc);
      remap_inode(&k_base->obj_id_and_type, true);
      btree_node_phys_t *child = BLOCKREF(omap_lookup(fs_object_map, child_loc).ov_paddr, OBJECT_TYPE_BTREE_NODE);
      process_fstree_node(child, tree_root);
    }
  }

  // TODO re-sort? so far it just works because our inode is the last entry anyway...

  tree_node->btn_o.checksum_number = apfs_checksum_postprocess(fletcher64((char*)tree_node+MAX_CKSUM_SIZE, sb->nx_block_size-MAX_CKSUM_SIZE));
  printf("<process_fstree_node() exit>\n");
}

int main(int argc, char **argv) {
  if (argc != 2) {
    printf("usage: echo {<old_inode> <new_inode>}+ | %s <image_path>\n", argv[0]);
    return 1;
  }
  char *image_path = argv[1];

  int image_fd = open(image_path, O_RDWR);
  if (image_fd == -1) err(1, "open image");
  struct stat image_st;
  if (fstat(image_fd, &image_st)) err(1, "fstat image");
  img_size = image_st.st_size;
  img = mmap(NULL, img_size, PROT_READ|PROT_WRITE, MAP_SHARED, image_fd, 0);
  if (img == MAP_FAILED) err(1, "mmap image");

  sb = BLOCKREF(0, OBJECT_TYPE_NX_SUPERBLOCK);
  printf("SB magic=0x%x xid=0x%"PRIx64" nx_omap_oid=0x%"PRIx64"\n", sb->nx_magic, sb->nx_o.o_xid, sb->nx_omap_oid);

  inode_pairs = 0;
  ino_map = calloc(MAX_MAPPINGS, sizeof(struct ino_mapping));
  uint64_t biggest_new_oid = 0;
  while (1) {
    char line[1000];
    if (fgets(line, sizeof(line), stdin) == NULL)
      break;
    if (inode_pairs == MAX_MAPPINGS) errx(1, "too many lines");
    char *space = strchr(line, ' ');
    if (space == NULL) errx(1, "bad line: '%s'", line);
    ino_map[inode_pairs].old = strtoull(line, NULL, 0);
    ino_map[inode_pairs].new = strtoull(space+1, NULL, 0);
    printf("parsed mapping 0x%llx -> 0x%llx\n", ino_map[inode_pairs].old, ino_map[inode_pairs].new);
    //if (ino_map[i].old >= sb->nx_next_oid) errx(1, "old oid too big");
    if (ino_map[inode_pairs].new < sb->nx_next_oid) errx(1, "new oid too small");
    if (ino_map[inode_pairs].new > biggest_new_oid) biggest_new_oid = ino_map[inode_pairs].new;
    if (inode_pairs != 0) {
      if (ino_map[inode_pairs-1].old >= ino_map[inode_pairs].old) errx(1, "input sorting is bogus (old)");
      if (ino_map[inode_pairs-1].new >= ino_map[inode_pairs].new) errx(1, "input sorting is bogus (new)");
    }
    inode_pairs++;
  }
  printf("got %d mappings\n", inode_pairs);

  // update sb->nx_next_oid and keep sb checksum okay
  if (sb->nx_o.checksum_number != apfs_checksum_postprocess(fletcher64((char*)sb+MAX_CKSUM_SIZE, sb->nx_block_size-MAX_CKSUM_SIZE)))
    errx(1, "root has bad checksum");
  /*
  sb->nx_next_oid = biggest_new_oid + 1;
  sb->nx_o.checksum_number = apfs_checksum_postprocess(fletcher64((char*)sb+MAX_CKSUM_SIZE, sb->nx_block_size-MAX_CKSUM_SIZE));
  */

/*
  printf("nx_xp_desc: base=0x%"PRIx64", blocks=0x%x\n", sb->nx_xp_desc_base, sb->nx_xp_desc_blocks);
  for (int i=0; i<sb->nx_xp_desc_blocks; i++) {
    obj_phys_t *obj_head = BLOCKREF(sb->nx_xp_desc_base + i, OBJECT_TYPE_INVALID);
    if ((obj_head->o_type & OBJECT_TYPE_MASK) == OBJECT_TYPE_NX_SUPERBLOCK) {
      nx_superblock_t *second_sb = (void*)obj_head;
      printf("superblock magic=0x%x xid=0x%"PRIx64" nx_omap_oid=0x%"PRIx64"\n", second_sb->nx_magic, second_sb->nx_o.o_xid, second_sb->nx_omap_oid);
    } else if ((obj_head->o_type & OBJECT_TYPE_MASK) == OBJECT_TYPE_CHECKPOINT_MAP) {
      //checkpoint_map_phys_t *chkp_map = (void*)obj_head;
      printf("checkpoint map\n");
    } else {
      errx(1, "unexpected type 0x%x in nx_xp_desc_base area, index %d", obj_head->o_type, i);
    }
  }
  puts("");
*/

  if (sb->nx_magic != NX_MAGIC) errx(1, "bad magic in sb");
  if (img_size % sb->nx_block_size != 0) errx(1, "image size not multiple of blocksize");
  if (img_size / sb->nx_block_size < sb->nx_block_count) errx(1, "image too small");
  oid_t fs_oid = 0;
  for (int i=0; i<NX_MAX_FILE_SYSTEMS; i++) {
    if (sb->nx_fs_oid[i] != 0) {
      if (fs_oid != 0) errx(1, "image contains more than one filesystem");
      fs_oid = sb->nx_fs_oid[i];
    }
  }
  if (fs_oid == 0) errx(1, "image contains no filesystem");
  printf("filesystem OID: 0x%"PRIx64"\n", fs_oid);
  sb_object_map = BLOCKREF(sb->nx_omap_oid, OBJECT_TYPE_OMAP);
  /*
  btree_node_phys_t *sb_btree_root = BLOCKREF(sb_object_map->om_tree_oid, OBJECT_TYPE_BTREE);
  btree_info_t *sb_btree_info = btree_info_by_root(sb_btree_root);
  printf("sb btree: flags=0x%x, nodesize=0x%x, keysize=0x%x, valsize=0x%x\n", sb_btree_info->bt_fixed.bt_flags, sb_btree_info->bt_fixed.bt_node_size, sb_btree_info->bt_fixed.bt_key_size, sb_btree_info->bt_fixed.bt_val_size);
  printf("sb btree root: flags=0x%hx, nkeys=0x%x, table_space.off=0x%hx, table_space.len=0x%hx\n", sb_btree_root->btn_flags, sb_btree_root->btn_nkeys, sb_btree_root->btn_table_space.off, sb_btree_root->btn_table_space.len);
  if (sb_btree_root->btn_table_space.off != 0) errx(1, "bogus btree");
  for (int i=0; i<sb_btree_root->btn_nkeys; i++) {
    struct kv_bufs bufs = btree_node_get_kv_bufs(i, sb_btree_root, sb_btree_root);
    printf("key/value: %p:0x%lx, %p:0x%lx\n", bufs.k.p, bufs.k.l, bufs.v.p, bufs.v.l);
  }
  */
  apfs_superblock_t *fs = BLOCKREF(omap_lookup(sb_object_map, fs_oid).ov_paddr, OBJECT_TYPE_FS);
  fs_object_map = BLOCKREF(fs->apfs_omap_oid, OBJECT_TYPE_OMAP);
  btree_node_phys_t *fstree = BLOCKREF(omap_lookup(fs_object_map, fs->apfs_root_tree_oid).ov_paddr, OBJECT_TYPE_BTREE);
  //omap_lookup(fstree, 1337);
  //tree_btree_root->btn_o.o_subtype == OBJECT_TYPE_FSTREE

  process_fstree_node(fstree, fstree);

  return 0;
}
